from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder
from .ModelValidator import ModelValidator
from .MCMCFitter import MCMCFitter
from .SummaryPlotter import SummaryPlotter

from . import helper



from pathlib import Path

import matplotlib.pyplot as plt

import numpy as np

import emcee
import corner
from matplotlib import rc

rc("font", **{"family": "sans-serif", "sans-serif": ["Helvetica"]})
rc("text")

from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder
from .MCMCFitter import MCMCFitter


class SummaryPlotter:
    def __init__(
        self,
        config: ConfigReader,
        pb: PredictionBuilder,
        fitter: MCMCFitter,
        **kwargs,
    ):
        """
        Create basic plots using results from MCMCFitter

        :param config: Config for the analysis
        :type config: ConfigReader

        :param pb: Model for predicting coefficient values
        :type pb: PredictionBuilder

        :param fitter: Fitter for the analysis
        :type fitter: MCMCFitter

        :param result_path: Path where plots should be saved
        :type result_path: str or pathlib.Path
        """

        if "result_path" not in kwargs:
            # make results directory based on run name
            run_name = config.params["config"]["run_name"]
            results_path = Path("results") / run_name
        else:
            results_path = kwargs["results_path"]
        results_path.mkdir(parents=True, exist_ok=True)

        self.path = results_path
        self.mcmc_params = np.mean(fitter.sampler.flatchain, axis=0)
        self.mcmc_params_cov = np.cov(np.transpose(fitter.sampler.flatchain))
        self.samples = fitter.sampler.chain.reshape(-1, len(config.prior_limits))
        self.config = config
        self.pb = pb

    def fit_result(self, ylabel: str, show_plot: bool = True, **kwargs):
        """
        Generate plot comparing the model with optimised coefficients with data present in the configuration within ConfigBuilder

        :param ylabel: Label for y-axis of plot
        :type ylabel: str

        :param show_plot: Determines whether `plt.show()` is called after a plot is created
        :type show_plot: boolean

        :param filename: Name of output file
        :type filename: str
        """

        if "filename" in kwargs:
            filename = kwargs["filename"]
        else:$
            filename = f"{self.config.run_name}_bestfit_predictions.png"

        data_label = f"Data ({self.config.run_name})"
        xlabel = self.config.observable if "xlabel" not in kwargs else kwargs["xlabel"]

        if len(self.config.coefficients) == 1:
            label_bestfit = f"Best Fit: {self.config.tex_labels[0]}={self.mcmc_params[0]:.3f}$\pm${np.sqrt(self.mcmc_params_cov):.3f}"
        else:
            label_bestfit = "Best Fit:"
            for c_name, c_value, c_error in zip(
                self.config.tex_labels,
                self.mcmc_params,
                np.sqrt(np.diagonal(self.mcmc_params_cov)),
            ):
                label_bestfit += f" {c_name}={c_value:.3f}$\pm${c_error:.3f}"


        data_err = np.diagonal(self.config.cov)

        half_bin_width = [
            (self.config.bins[i + 1] - self.config.bins[i]) / 2
            for i in range(len(self.config.bins) - 1)
        ]
        plt.errorbar(
            self.config.x_vals,
            self.config.data,
            fmt=".",
            xerr=half_bin_width,
            yerr=data_err,
            label=data_label,
        )

        plt.errorbar(
            self.config.x_vals,
            self.pb.make_prediction(self.mcmc_params),
            fmt="mo",
            ls="None",
            xerr=0.0,
            yerr=0.0,
            label=label_bestfit,
        )

        ax = plt.gca()
        ax.set_xlabel(xlabel, fontsize=18)
        ax.set_ylabel(ylabel, fontsize=18)
        plt.legend(loc=2)
        plt.savefig(self.path / filename)

        if show_plot:
            plt.show()
        plt.close()

    def corner(self, show_plot: bool = True, **kwargs):
        """
        Generate corner plot using `corner <https://corner.readthedocs.io/en/latest/>`_.

        :param show_plot: Determines whether `plt.show()` is called after a plot is created.
        :type show_plot: bool

        :param filename: Provide filename for output file
        :type filename: str
        """
        if "filename" in kwargs:
            filename = kwargs["filename"]
        else:
            filename = f"{self.config.run_name}.png"

        #  make corner plot
        labels = []
        ranges = []
        for c in self.config.params["config"]["model"]["prior_limits"].keys():
            label = "$" + c + "$"
            labels.append(label)
            ranges.append(self.config.params["config"]["model"]["prior_limits"][c])

        fig = corner.corner(
            self.samples,
            labels=self.config.tex_labels,
            label_kwargs={"fontsize": 18},
            range=ranges,
            quantiles=[0.16, 0.84],
            levels=(1 - np.exp(-0.5),),
            truths=np.zeros(len(labels)),
            show_titles=True,
            title_kwargs={"fontsize": 18},
        )

        ax0 = fig.add_subplot(999)
        ax0.axis("off")

        fig.savefig(self.path / filename)
        if show_plot:
            plt.show()
        plt.close()


# class SummaryPlotter:
# def __init__(
# self,
# config: ConfigReader,
# pb: PredictionBuilder,
# sampler: emcee.EnsembleSampler,
# ):

# samples = sampler.chain.reshape(-1, len(config.prior_limits))

# # make directory to hold results of this run
# run_name = config.params["config"]["run_name"]
# results_path = Path("results") / run_name
# results_path.mkdir(parents=True, exist_ok=True)

# mcmc_params = np.mean(sampler.flatchain, axis=0)
# mcmc_params_cov = np.cov(np.transpose(sampler.flatchain))

# data_label = f"Data ({run_name})"
# max_val = (1.5) * (max(config.params["config"]["data"]["central_values"]))
# min_val = (0.0) * (min(config.params["config"]["data"]["central_values"]))
# xlabel = config.observable
# ylabel = "d$\sigma_{tt}$/dX"

# # make plots of prediction at max point of posterior versus data
# coefficients = list(config.coefficients)
# pl.figure()
# label_string_bestfit = "best-fit: ("
# for c in range(0, len(config.coefficients)):
# if c == (len(config.coefficients) - 1):
# label_string_bestfit = (
# label_string_bestfit
# + coefficients[c]
# + " = "
# + "%.1f" % mcmc_params[c]
# + ")"
# )
# else:
# label_string_bestfit = (
# label_string_bestfit
# + coefficients[c]
# + " = "
# + "%.1f" % mcmc_params[c]
# + ", "
# )

# pl.errorbar(
# config.x_vals,
# pb.make_prediction(mcmc_params),
# fmt="m",
# xerr=0.0,
# yerr=0.0,
# label=label_string_bestfit,
# )
# pl.errorbar(
# config.x_vals,
# config.params["config"]["data"]["central_values"],
# fmt="o",
# xerr=0.25,
# yerr=0.05,
# label=data_label,
# )
# pl.axis(
# [
# config.x_vals[0] - 0.25,
# config.x_vals[len(config.x_vals) - 1] + 0.25,
# min_val,
# max_val,
# ]
# )
# ax = pl.gca()
# labely = ax.set_xlabel(xlabel, fontsize=18)
# labely = ax.set_ylabel(ylabel, fontsize=18)
# ax.xaxis.set_label_coords(0.85, -0.065)
# ax.yaxis.set_label_coords(-0.037, 0.83)
# pl.legend(loc=2)
# pl.savefig(results_path / f"{run_name}_bestfit_predictions.png")
# pl.close()

# #  make "corner" plot
# labels = []
# ranges = []
# for c in config.params["config"]["model"]["prior_limits"].keys():
# label = "$" + c + "$"
# labels.append(label)
# ranges.append(config.params["config"]["model"]["prior_limits"][c])

# fig = corner.corner(
# samples,
# labels=config.tex_labels,
# label_kwargs={"fontsize": 18},
# range=ranges,
# quantiles=[0.16, 0.84],
# levels=(1 - np.exp(-0.5),),
# truths=np.zeros(len(labels)),
# show_titles=True,
# title_kwargs={"fontsize": 18},
# )

# plotfilename = results_path / f"{run_name}.png"

# ax0 = fig.add_subplot(999)
# ax0.axis("off")

# fig.savefig(plotfilename)
# pl.close()

# bestFits = []
# marginUncsUp = []
# marginUncsDown = []
# x = []

# for i in range(len(labels)):
# mcmc = np.percentile(samples[:, i], [16, 50, 84])
# q = np.diff(mcmc)
# txt = "\mathrm{{{3}}} = {0:.3f}_{{-{1:.3f}}}^{{{2:.3f}}}"
# txt = txt.format(mcmc[1], q[0], q[1], labels[i])
# x.append(i)
# bestFits.append(mcmc[1])
# marginUncsUp.append(q[0])
# marginUncsDown.append(q[1])

# pl.figure()
# pl.tight_layout()
# pl.gcf().subplots_adjust(bottom=0.15)

# fig = pl.errorbar(
# x,
# bestFits,
# yerr=[marginUncsDown, marginUncsUp],
# fmt="o",
# label=r"posterior median and CI",
# )
# pl.xticks(range(len(labels)), config.tex_labels, rotation="45", fontsize=21)
# ax = pl.gca()
# ax.yaxis.set_label_coords(-0.075, 0.83)
# labely = ax.set_ylabel(r"$c_{i}$ [$GeV^{-2}$]", fontsize=18)
# pl.legend(loc=1, fontsize=14)
# pl.savefig(results_path / f"{run_name}_bestFits.png")
# pl.close()

# ############################################
# ######## Corner with 1-d CL overlaid  ######
# ############################################

# fig_overlay = corner.corner(
# samples,
# labels=config.tex_labels,
# label_kwargs={"fontsize": 21},
# range=ranges,
# color="k",
# quantiles=[0.16, 0.84],
# levels=(1 - np.exp(-0.5),),
# truths=np.zeros(len(labels)),
# show_titles=True,
# title_kwargs={"fontsize": 19},
# hist2d_kwargs={"fill_contours": True, "plot_density": True},
# )

# resplot = image.imread(results_path / f"{run_name}_bestFits.png")

# ax0 = fig_overlay.add_subplot(322)
# ax0.axis("off")
# img = ax0.imshow(resplot)

# ax0 = fig_overlay.add_subplot(5, 5, 15)
# ax0.axis("off")

# plotfilename = results_path / f"{run_name}_overlay.png"

# fig_overlay.savefig(plotfilename)
# pl.close()

# # make plots of best fit prediction versus data
# pl.figure()
# label_string_bestfit = "best-fit prediction"

# pl.errorbar(
# config.x_vals,
# pb.make_prediction(bestFits),
# fmt="m",
# xerr=0.0,
# yerr=0.0,
# label=label_string_bestfit,
# )
# pl.errorbar(
# config.x_vals,
# config.params["config"]["data"]["central_values"],
# fmt="o",
# xerr=0.25,
# yerr=np.sqrt(
# np.diagonal(config.params["config"]["data"]["covariance_matrix"])
# ),
# label=data_label,
# )
# pl.axis(
# [
# config.x_vals[0] - 0.25,
# config.x_vals[len(config.x_vals) - 1] + 0.25,
# min_val,
# max_val,
# ]
# )
# ax = pl.gca()
# labely = ax.set_xlabel(xlabel, fontsize=18)
# labely = ax.set_ylabel(ylabel, fontsize=18)
# ax.xaxis.set_label_coords(0.85, -0.065)
# ax.yaxis.set_label_coords(-0.037, 0.83)
# pl.legend(loc=1)
# pl.savefig(results_path / f"{run_name}_bestfit_predictions.png")
# pl.close()

# # make fit summary json
# fitSummary = {}
# fitSummary["bestFit"] = bestFits
# fitSummary["bestFit"] = bestFits
# fitSummary["UncsUp"] = marginUncsUp
# fitSummary["UncsDown"] = marginUncsDown
# fitSummary["labels"] = labels
# fitSummary["x"] = x

# with open(run_name + ".json", "w") as fs:
# json.dump(fitSummary, fs)

# pl.figure()

# # best-fit prediction and random samples compared to data
# uncX = np.absolute((np.diff(config.bins))) / 2.0
# dataUncY = np.sqrt(np.diag(config.cov))
# pl.errorbar(
# config.x_vals,
# config.params["config"]["data"]["central_values"],
# fmt="o",
# xerr=uncX,
# yerr=dataUncY,
# label=data_label,
# )

# inds = np.random.randint(len(samples), size=499)
# for ind in inds:
# sample = samples[ind]
# pl.plot(config.x_vals, pb.make_prediction(sample), "C1", alpha=0.02)

# pl.plot(
# config.x_vals,
# pb.make_prediction(samples[42]),
# "C1",
# label="500 random samples",
# alpha=0.1,
# )

# ax = pl.gca()
# labelx = ax.set_xlabel(xlabel, fontsize=18)
# labely = ax.set_ylabel(ylabel, fontsize=18)
# ax.xaxis.set_label_coords(0.85, -0.065)
# ax.yaxis.set_label_coords(-0.037, 0.83)
# pl.legend(loc=1)
# pl.savefig(results_path / f"{run_name}_postfit_predictions.png")
# pl.close()

# pl.figure()
# fig, axes = pl.subplots(len(labels), figsize=(10, 7), sharex=True)
# samples = sampler.get_chain()
# for i in range(len(labels)):
# ax = axes[i]
# ax.plot(samples[:, :, i], "k", alpha=0.3)
# ax.set_xlim(0, len(samples))
# ax.set_ylabel(labels[i])
# ax.yaxis.set_label_coords(-0.1, 0.5)

# axes[-1].set_xlabel("step number")

# pl.savefig(results_path / f"{run_name}_walkersPaths.png")
# pl.close()

# # covariance matrix of coefficicents
# fig, ax = pl.subplots()
# im = ax.imshow(mcmc_params_cov, cmap=pl.cm.Blues)

# ax.set_xticks(np.arange(len(config.tex_labels)))
# ax.set_yticks(np.arange(len(config.tex_labels)))
# # ... and label them with the respective list entries
# ax.set_xticklabels(config.tex_labels)
# ax.set_yticklabels(config.tex_labels)

# # fixing yticks with matplotlib.ticker "FixedLocator"
# # ticks_loc = ax.get_yticks().tolist()
# # ticks_loc = config.tex_labels

# # ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
# # label_format = '{:,.0f}'

# # pl.imshow(mcmc_params_cov, cmap=pl.cm.Blues)
# # ax.set_xticks(np.arange(len(config.tex_labels)))

# # ax.set_yticklabels(ticks_loc)

# pl.savefig(results_path / f"{run_name}_mcmc_params_cov.png")
# pl.close()

# # Write SM pred to text file for validation
# sm_file_name = results_path / f"{run_name}_postFit_sm_pred.txt"
# f = open(sm_file_name, "w")
# sm_pred = str(
# repr(
# pb.make_prediction(
# np.zeros(len(config.params["config"]["model"]["prior_limits"]))
# )
# )
# )
# f.write(sm_pred)
# f.close()



import re as _re
from pathlib import Path as _Path
import numpy as _np


def convert_hwu_to_numpy(
    filename: _Path, num_of_hist: int = 1
) -> (_np.ndarray, _np.ndarray, _np.ndarray):
    """
    Convert a HwU (histogram with uncertainties) file from MadGraph into a JSON file which is compatible with dEFT.
    Only parses the first histogram in the file by default.
    """

    # Known Bug: Only considers the last histogram as the final bin and will only consider the right edge on the final bin

    with open(filename, "r") as hwu_f:
        bin_lefts = _np.empty(0)
        bin_rights = _np.empty(0)
        central_values = _np.empty(0)
        for i in range(num_of_hist):
            line = next(hwu_f)
            while (match := _re.search(r"<histogram> (\d+) ", line)) is None:
                line = next(hwu_f)
                if line is None:
                    raise RuntimeError(f"No histogram found in HwU file: {filename}")
            bins = int(match.group(1))

            line = next(hwu_f)
            for i in range(bins):
                # The 2-index in HwU are central value of bin
                values = line.strip().split()
                central_values = _np.append(central_values, float(values[2]))
                bin_lefts = _np.append(bin_lefts, float(values[0]))
                bin_rights = _np.append(bin_rights, float(values[1]))
                line = next(hwu_f)

    assert len(bin_lefts) == len(central_values)
    assert len(bin_rights) == len(central_values)
    return (bin_lefts, bin_rights, central_values)



import itertools
from typing import List

import numpy as np
from sklearn.linear_model import LinearRegression

from .ConfigReader import ConfigReader


def _triangular_number(n: int) -> int:
    return int(n * (n + 1) / 2)


class PredictionBuilder:
    """
    Constructs Linear Regression Model for the wilson coefficient of each operator for some observable
    """

    def __init__(
        self,
        config: ConfigReader,
    ):
        """
        Constructor for PredictionBuilder

        :param config: Configuration for analysis.
        :type config: ConfigReader.
        """
        self.nOps = len(config.prior_limits)
        self.nSamples = int((self.nOps + 1) * (self.nOps + 2) / 2)
        self.model = self._build_regression_model(config.samples, config.predictions)

    def _build_regression_model(
        self, samples: List[float], preds: List[List[float]]
    ) -> LinearRegression:
        """Initialise morphing model using samples with predicted values"""
        if len(preds) < self.nSamples:
            raise TypeError(
                "morphing with "
                + str(self.nOps)
                + " coefficients requires at least "
                + str(self.nSamples)
                + " samples,  but only "
                + str(len(preds))
                + " are provided"
            )

        # convert to vector of coefficient factors to linearise the morphing
        cInputAr = self._make_coefficients(samples)

        # define model
        model = LinearRegression()

        # fit model
        model.fit(cInputAr, preds)

        return model

    def _make_coefficients(self, ci: List[float]) -> np.ndarray:
        X = np.array([])
        num_rows = np.shape(ci)[0]
        for row in ci:
            # Account for quadratic self term
            X = np.append(X, row ** 2)

            # Account for quadratic cross term
            combs = itertools.combinations(list(row), 2)
            for comb in combs:
                X = np.append(X, comb[0] * comb[1])
        X = X.reshape(num_rows, _triangular_number(len(ci[0])))
        return X

    def _make_coeff_point(self, ci: np.ndarray) -> np.ndarray:
        X = np.array([])
        X = np.append(X, ci ** 2)
        combs = itertools.combinations(list(ci), 2)
        for comb in combs:
            X = np.append(X, comb[0] * comb[1])

        X = X.reshape(1, _triangular_number(len(ci)))
        return X

    def make_prediction(self, c: np.ndarray) -> np.ndarray:
        """
        Produce the predicted observable for a set of coefficients (excluding SM coefficient)

        :param c: A set of coefficients to be used by the model in order to make a prediction. Should have a coefficient for each operator in the model.
        :type c: numpy.ndarray
        """

        if len(c) != self.nOps:
            raise Exception(
                f"An incorrect number of coefficients were supplied. Model requires coeffiecient for {self.nOps} operators but {len(c)} were supplied"
            )

        c = np.append(1.0, c)
        cInputAr = self._make_coeff_point(c)
        pred = self.model.predict(cInputAr)
        return pred[0]



import sys
import json
import numpy as np
from numpy.linalg import inv


class ConfigReader:
    """
    Reads JSON configuration file and acts as a full configuration of the analysis

    :param run_name: Label for analysis
    :type run_name: str

    :param observable: Label for observables
    :type observable: str, list[str]

    :param bins: Bin edges
    :type bins: list[float], list[list[float]]

    :param data: Central value of bins
    :type data: list[float], list[list[float]]

    :param coefficients: Label for each coefficient

    :param prior_limits: Bounds for each coefficient.
    :type prior_limits: list[float]

    :param samples: Samples of coefficient values
    :type samples: list[list[float]]

    :param predictions: Predicted Monte Carlo signal for the corresponding set of coefficients
    :type predictions: list[list[float]]

    :param inclusive_k_factor: K factor used to scale predictions depending on order of Monte Carlo generation
    :type inclusive_k_factor: float

    :param n_walkers: Number of walkers to use for MCMC fitting
    :type n_walkers: int

    :param n_burnin: Number of burn in runs to perform before beginning MCMC fitting
    :type n_burnin: int

    :param n_total: Number of total iterations for MCMC fitting
    :type n_total: int
    """

    # TODO: Sort out parameters and improve usability

    def __init__(self, filename: str):
        """
        Constructor for ConfigReader

        :param filename: Configuration file
        :type filename: str
        """
        self.filename = filename
        with open(filename, "r") as f:
            config = json.load(f)

        try:
            self.params = config
            self.run_name = self.params["config"]["run_name"]
            self.observable = self.params["config"]["data"]["observable"]
            self.bins = self.params["config"]["data"]["bins"]
            self.data = self.params["config"]["data"]["central_values"]
            self.cov = self.params["config"]["data"]["covariance_matrix"]

            if len(self.cov) != len(self.data):
                raise Exception(
                    "Covariance matrix does not have suitable dimensions. Expected ({len(self.data), len(self.data)})"
                )

            for row in self.cov:
                if len(row) != len(self.data):
                    raise Exception(
                        "Covariance matrix does not have suitable dimensions. Expected ({len(self.data), len(self.data)})"
                    )

            self.prior_limits = self.params["config"]["model"]["prior_limits"]
            for label, limits in self.prior_limits.items():
                if len(limits) != 2:
                    raise Exception(f"Coefficient {label} does not have 2 values.")
                if limits[0] >= limits[1]:
                    raise Exception(
                        f"Prior limit for coefficient {label} are invalid: {limit}"
                    )

            self.coefficients = list(
                self.params["config"]["model"]["prior_limits"].keys()
            )
            self.n_walkers = self.params["config"]["fit"]["n_walkers"]
            self.n_burnin = self.params["config"]["fit"]["n_burnin"]
            self.n_total = self.params["config"]["fit"]["n_total"]
            self.samples = np.array(self.params["config"]["model"]["samples"])
            self.inclusive_k_factor = self.params["config"]["model"][
                "inclusive_k_factor"
            ]

            if (
                "input" not in self.params["config"]["model"]
                or self.params["config"]["model"]["input"] == "numpy"
            ):
                self.predictions = np.array(
                    self.params["config"]["model"]["predictions"]
                )
                if len(self.predictions) != len(self.samples):
                    raise Exception(
                        "Number of samples must equal number of predictions"
                    )
        except KeyError as err:
            print(
                f"Error reading file {filename}: Could not find option {err}",
                file=sys.stderr,
            )
            exit(0)

        self.tex_labels = ["$" + c + "$" for c in self.coefficients]

        x_vals = np.zeros(len(self.bins) - 1)
        for x_val in range(0, len(self.bins) - 1):
            x_vals[x_val] = (self.bins[x_val] + self.bins[x_val + 1]) / 2.0
        self.x_vals = x_vals



from multiprocessing import Pool
from typing import Optional

import numpy as np
import emcee

from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder

_PB: Optional[PredictionBuilder] = None
_DATA: Optional[np.ndarray] = None
_ICOV: Optional[np.ndarray]  = None
_BOUNDS: Optional[np.ndarray]  = None


def ln_prior(c: np.ndarray) -> float:
    lnp = 0.0

    if (c < _BOUNDS[0]).any() or (_BOUNDS[1] < c).any():
        lnp = -np.inf
    return lnp


def ln_prob(
    c: np.ndarray,
) -> float:
    pred = _PB.make_prediction(c)
    diff = pred - _DATA
    ll = (-np.dot(diff, np.dot(_ICOV, diff))) + ln_prior(c)
    return ll


class MCMCFitter:
    """ Use Markov chain Monte Carlo method to fit the model from the PredictionBuilder to some data specified in the configuration file """

    def __init__(
        self,
        config: ConfigReader,
        pb: PredictionBuilder,
        initial_variance: float = 1e-4,
        use_multiprocessing: bool = False,
    ):
        n_walkers = config.n_walkers

        n_dim = int(len(config.prior_limits))
        n_burnin = config.n_burnin
        n_total = config.n_total
        p0 = [
            initial_variance * np.random.randn(n_dim) for _ in range(n_walkers)
        ]  # Initial position of walkers

        global _DATA
        global _ICOV
        global _BOUNDS
        global _PB
        _DATA = config.data
        _ICOV = np.linalg.inv(config.cov)
        _BOUNDS = np.array(list(config.prior_limits.values())).T
        _PB = pb

        if use_multiprocessing:
            with Pool() as pool:
                sampler = emcee.EnsembleSampler(
                    n_walkers,
                    n_dim,
                    ln_prob,
                    pool=pool,
                )
                # Run burn in runs
                pos, _, _ = sampler.run_mcmc(p0, n_burnin, progress=True)
                sampler.reset()

                # Perform proper run
                sampler.run_mcmc(pos, n_total, progress=True)
        else:
            sampler = emcee.EnsembleSampler(
                n_walkers,
                n_dim,
                ln_prob,
            )
            # Run burn in runs
            pos, _, _ = sampler.run_mcmc(p0, n_burnin, progress=True)
            sampler.reset()

            # Perform proper run
            sampler.run_mcmc(pos, n_total, progress=True)

        self.sampler = sampler



from pathlib import Path
from typing import List

import numpy as np
import matplotlib.pyplot as plt

from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder


class ModelValidator:
    """
    Allows for a model to be validated against dataset with known operator paramters (like MC signal). The model must be defined using `PredictionBuilder` class.
    """

    def __init__(self, pb: PredictionBuilder):
        self.nOps = pb.nOps
        self.pb = pb

    def validate(self, config_test: ConfigReader) -> (List[float], List[np.ndarray]):
        """ Generate predictions for the test samples in the test configuration using the model from the PredictionBuilder """

        test_samples = config_test.params["config"]["model"]["samples"]
        test_preds = np.asarray(config_test.params["config"]["model"]["predictions"])
        assert self.nOps == len(
            config_test.params["config"]["model"]["prior_limits"]
        ), f"Operator mismatch ({self.nOps} : {len(config_test.params['config']['model']['prior_limits'])}). Make sure the supplied ConfigReader has the correct number of operators for the model"

        print(
            "##############################################################\n"
            "###############  VALIDATION OF MORPHING MODEL  ###############\n"
            "##############################################################\n"
            "\n"
            f"Number of operators: {self.nOps}\n"
            f"Number of validation cases: {len(test_samples)}\n"
        )

        # Generate predictions using model
        model_preds = [None] * len(test_samples)
        for i, sample in enumerate(test_samples):
            model_preds[i] = self.pb.make_prediction(sample[1:])
        assert len(model_preds[0]) == len(test_preds[0])

        diff_preds = np.square(model_preds - test_preds).sum(axis=0)
        print(f"Total Square difference: {diff_preds}")

        return test_samples, model_preds

    def comparison_plot(
        self,
        config: ConfigReader,
        model_predictions: List[np.ndarray],
        filepath: str = None,
        use_pdf: bool = False,
    ):
        """ Produces a comparison plots for each test case at a specified filepath or at `./results/{config.run_name}_validation`. The supplied ConfigReader should be related to the validation configuration."""
        run_name = config.run_name
        results_path = (
            Path("results") / f"{run_name}_validation" if not filepath else filepath
        )
        results_path.mkdir(parents=True, exist_ok=True)

        # Plot comparison between the model and the monte carlo generated plots
        for i, (m_pred, t_pred) in enumerate(
            zip(model_predictions, config.predictions)
        ):
            plt.scatter(config.x_vals, m_pred, label="Model")
            plt.scatter(config.x_vals, t_pred, label="MC")
            plt.xlabel(f"{config.observable}")
            plt.legend()
            plt.savefig(
                results_path / f"model_comparison_{i}.{'pdf' if use_pdf else 'png'}"
            )
            plt.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deft_hep']

package_data = \
{'': ['*']}

install_requires = \
['corner>=2.2.1,<3.0.0',
 'emcee>=3.0.2,<4.0.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.1,<2.0.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'tqdm>=4.59.0,<5.0.0']

setup_kwargs = {
    'name': 'deft-hep',
    'version': '0.5.1',
    'description': 'dEFT is a tool for performing fits of EFT coefficients to HEP data in seconds',
    'long_description': None,
    'author': 'Alexander Veltman',
    'author_email': 'alexveltman@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/codecalec/dEFT',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
